! TTT -- an implementation of Tic Tac Toe
! -- Object-oriented dual game version --
! Copyright (C) 2003 Nils M Holm.  All rights reserved.
! See the file LICENSE of the T3X distribution
! for conditions of use.

class ttt_oo(t3x, ttyctl)

 object	t[t3x], tty[ttyctl];

 var	Field[9];	! Playing field
 var	Moves;		! Move database
 var	Xpos, Ypos;	! Current position
 var	Over;		! Game over flag

 ! Special return codes
 public const	GAME_QUIT = -1,	! Player quit
		GAME_OVER = -2,	! Game over
		GAME_SKIP = -3;	! Player skipped move

 ! Draw empty field
 drawframe() do
	tty.move(Xpos, Ypos+0); tty.writes("   |   |   ");
	tty.move(Xpos, Ypos+1); tty.writes("---+---+---");
	tty.move(Xpos, Ypos+2); tty.writes("   |   |   ");
	tty.move(Xpos, Ypos+3); tty.writes("---+---+---");
	tty.move(Xpos, Ypos+4); tty.writes("   |   |   ");
 end

 public init(x, y) do
	Xpos := x;
	Ypos := y;
	Over := 0;
	Moves := [
		[ [0,2,2], [2,2,2] ],
		[ [2,0,2], [2,2,2] ],
		[ [2,2,0], [2,2,2] ],
		[ [0,1,1], [2,1,1] ],
		[ [1,0,1], [1,2,1] ],
		[ [1,1,0], [1,1,2] ],
		[ [0,1,0], [2,1,0] ],
		[ [0,0,1], [0,2,1] ],
		[ [1,0,0], [1,2,0] ],
		[ [0,1,2], [2,1,2] ],
		[ [0,2,1], [2,2,1] ],
		[ [1,0,2], [1,2,2] ],
		[ [1,2,0], [1,2,2] ],
		[ [2,0,1], [2,2,1] ],
		[ [2,1,0], [2,1,2] ],
		[ [0,0,2], [2,0,2] ],
		[ [0,2,0], [2,2,0] ],
		[ [2,0,0], [2,0,2] ],
		[ [0,0,0], [0,2,0] ],
		0
	];
	t.memfill(Field, 0, 9*t.bpw());		! Clear playing field
	tty.init();
	drawframe();
 end

 ! Finish game
 public fini() do
	if (Over) return;	! Already over?
	tty.fini();
	Over := 1;
 end

 ! Draw playing field with pieces
 drawfield() do var i, j;
	for (j=0, 3) do
		for (i=0, 3) do
			tty.move(Xpos+1+4*i, Ypos+2*j);
			tty.writec(Field[j*3+i] = 1 -> 'X':
				Field[j*3+i] = 2 -> 'O': '\s');
		end
	end
 end

 ! Match move in fp0,fp1,fp2 against a rule of the move database.
 ! Perform associated move in case of a match.
 match(fp0, fp1, fp2, rule) do var p, q;
	p := rule[0];
	q := rule[1];
	if (fp0[0] = p[0] /\ fp1[0] = p[1] /\ fp2[0] = p[2]) do
		fp0[0] := q[0];
		fp1[0] := q[1];
		fp2[0] := q[2];
		return 1;
	end
 end

 ! Do a move
 public move(n) do var i;
	Field[n] := 1;		! Place human player's piece
	if (Field[4] = 0) do	! Allocate center, if unoccupied
		Field[4] := 2;
		drawfield();
		return;
	end
	! Find best move
	i := 0;
	while (Moves[i]) do
		if (match(@Field[0], @Field[4], @Field[8], Moves[i])) leave;
		if (match(@Field[2], @Field[4], @Field[6], Moves[i])) leave;
		if (match(@Field[0], @Field[1], @Field[2], Moves[i])) leave;
		if (match(@Field[3], @Field[4], @Field[5], Moves[i])) leave;
		if (match(@Field[6], @Field[7], @Field[8], Moves[i])) leave;
		if (match(@Field[0], @Field[3], @Field[6], Moves[i])) leave;
		if (match(@Field[1], @Field[4], @Field[7], Moves[i])) leave;
		if (match(@Field[2], @Field[5], @Field[8], Moves[i])) leave;
		i := i+1;
	end
	drawfield();
	! Made a winning move?
	if (i < 3) do
		tty.move(Xpos, Ypos+6);
		tty.writes("Oops!");
		self.fini();
		return;
	end
 end

 ! Let the human player select a cell interactively
 public select() do var i, j, k;
	if (Over) return GAME_SKIP;		! Game already over?
	for (i=0, 9) if (\Field[i]) leave;	! Find free cell
	if (i = 9) do				! No free cells?
		self.fini();			!  Game ended in a draw
		return GAME_SKIP;
	end
	j := i / 3;	! y coordinate
	i := i mod 3;	! x coordinate
	tty.mode(1);
	while (1) do
		! Place cursor and accept key
		tty.move(Xpos+1+4*i, Ypos+2*j);
		k := tty.readc();
		! Move cursor
		ie (k = 'l' \/ k = ttyctl.K_RIGHT) do
			i := i+1;
			if (i > 2) i := 0;
		end
		else ie (k = 'h' \/ k = ttyctl.K_LEFT) do
			i := i-1;
			if (i < 0) i := 2;
		end
		else ie (k = 'j' \/ k = ttyctl.K_DOWN) do
			j := j+1;
			if (j > 2) j := 0;
		end
		else ie (k = 'k' \/ k = ttyctl.K_UP) do
			j := j-1;
			if (j < 0) j := 2;
		end
		! Cell selected
		else ie (k = '\r') do
			ie (Field[j*3+i])
				tty.writec('\a');	! BEEP - not empty
			else
				leave;			! OK
		end
		! Abort game
		else ie (k = ttyctl.K_ESC \/ k = 'q') do
			self.fini();
			return GAME_QUIT;
		end
		! Skip move
		else if (k = '\t') do
			return GAME_SKIP;
		end
	end
	tty.mode(0);
	return j*3+i;	! Return selected cell
 end

 public isover() return Over;

end

module ttt_games(ttyctl, ttt_oo);

object	tty[ttyctl],
	game1[ttt_oo], game2[ttt_oo];	! Instantiate games

do var i, g, r;
	tty.init();
	tty.clear();
	game1.init(8, 4);
	game2.init(24, 4);
	g := 0;			! Select game 1
	while (1) do
		if (game1.isover() /\ game2.isover())
			leave;
		! Get return code
		r := (g-> game2.select(): game1.select());
		ie (r = TTT_OO.GAME_SKIP)	! Move skipped,
			g := \g;
		else ie (r = TTT_OO.GAME_QUIT)	! Player quit
			leave;
		else ie (g)			! Game2 cell selected
			game2.move(r);	
		else
			game1.move(r);		! Game1 cell selected
	end
	game2.fini();
	game1.fini();
	tty.move(0, tty.lines()-2);
	tty.fini();
end
